@@ -8,7 +8,7 @@
  *             Andrae Muys <andrae@humbug.org.au.au>
  *           - uses libbzip2 by Julian Seward (http://sources.redhat.com/bzip2/)
  *
- *	Date  : February 5, 2006
+ *	Date  : March 20, 2007
  *
  *
  *  Contributions
@@ -26,15 +26,19 @@
  *  Benjamin Reed <ranger@befunk.com> - autodetection for # of CPUs in OSX
  *  Chris Dearman <chris@mips.com> - fixed pthreads race condition
  *  Richard Russon <ntfs@flatcap.org> - help fix decompression bug
+ *  Paul Pluzhnikov <paul@parasoft.com> - fixed minor memory leak
+ *  Anbal Monsalve Salazar <anibal@debian.org> - creates and maintains Debian packages
+ *  Steve Christensen - creates and maintains Solaris packages (sunfreeware.com)
+ *  Alessio Cervellin - creates and maintains Solaris packages (blastwave.org)
  *
  *  Specials thanks for suggestions and testing:  Phillippe Welsh,
  *  James Terhune, Dru Lemley, Bryan Stillwell, George Chalissery,
  *  Kir Kolyshkin, Madhu Kangara, Mike Furr, Joergen Ramskov, Kurt Fitzner,
  *  Peter Cordes, Oliver Falk, Jindrich Novy, Benjamin Reed, Chris Dearman,
- *  and Richard Russon.
+ *  Richard Russon, Anbal Monsalve Salazar, Jim Leonard, Paul Pluzhnikov.
  *
  *
- * This program, "pbzip2" is copyright (C) 2003-2006 Jeff Gilchrist.
+ * This program, "pbzip2" is copyright (C) 2003-2007 Jeff Gilchrist.
  * All rights reserved.
  *
  * The library "libbzip2" which pbzip2 uses, is copyright
@@ -73,7 +77,7 @@
  *
  * Jeff Gilchrist, Ottawa, Canada.
  * pbzip2@compression.ca
- * pbzip2 version 0.9.6 of February 5, 2006
+ * pbzip2 version 1.0.1 of March 20, 2007
  *
  */
 #include <vector>
@@ -98,7 +102,9 @@
 #ifdef __APPLE__
 #include <sys/sysctl.h>
 #endif
-
+#ifdef __sun
+#include <sys/loadavg.h>
+#endif
 #ifndef  __BORLANDC__
 #define __STDC_FORMAT_MACROS
 #include <inttypes.h>
@@ -106,10 +112,21 @@
 #define PRIu64 "Lu"
 #define strncasecmp(x,y,z) strncmpi(x,y,z)
 #endif
+#ifdef __osf__
+#define PRIu64 "llu"
+#endif
 
 // uncomment for debug output
 //#define PBZIP_DEBUG
 
+// uncomment to disable load average code (may be required for some platforms)
+//#define PBZIP_NO_LOADAVG
+
+// detect systems that are known not to support load average code
+#if defined (WIN32) || defined (__CYGWIN32__) || defined (__MINGW32__) || defined (__BORLANDC__) || defined (__hpux) || defined (__osf__)
+	#define PBZIP_NO_LOADAVG
+#endif
+
 #ifdef WIN32
 #define usleep(x) Sleep(x/1000)
 #define LOW_DWORD(x)  ((*(LARGE_INTEGER *)&x).LowPart)
@@ -1127,12 +1144,10 @@ void queueDelete (queue *q)
 		q->notEmpty = NULL;
 	}
 
-	if (q->consumers != NULL)
-	{
-        delete q->consumers;
-		q->consumers = NULL;
-	}
-
+    delete [] q->consumers;
+    delete [] q->blockNum;
+    delete [] q->bufSize;
+    delete [] q->buf;
 	delete q;
 	q = NULL;
 
@@ -1390,9 +1405,8 @@ int writeFileMetaData(char *fileName)
  */
 void banner()
 {
-	fprintf(stderr, "Parallel BZIP2 v0.9.6 - by: Jeff Gilchrist [http://compression.ca]\n");
-	fprintf(stderr, "[Feb. 05, 2006]             (uses libbzip2 by Julian Seward)\n\n");
-	fprintf(stderr, "** This is a BETA version - Use at your own risk! **\n");
+	fprintf(stderr, "Parallel BZIP2 v1.0.1 - by: Jeff Gilchrist [http://compression.ca]\n");
+	fprintf(stderr, "[Mar. 20, 2007]             (uses libbzip2 by Julian Seward)\n");
 
 	return;
 }
@@ -1404,12 +1418,15 @@ void usage(char* progname)
 {
 	banner();
 	fprintf(stderr, "\nInvalid command line!  Aborting...\n\n");
-	fprintf(stderr, "Usage: %s [-1 .. -9] [-b#cdfkp#qrtV] <filename> <filename2> <filenameN>\n", progname);
+	fprintf(stderr, "Usage: %s [-1 .. -9] [-b#cdfklp#qrtV] <filename> <filename2> <filenameN>\n", progname);
 	fprintf(stderr, " -b#      : where # is the file block size in 100k (default 9 = 900k)\n");
 	fprintf(stderr, " -c       : output to standard out (stdout)\n");
 	fprintf(stderr, " -d       : decompress file\n");
 	fprintf(stderr, " -f       : force, overwrite existing output file\n");
 	fprintf(stderr, " -k       : keep input file, don't delete\n");
+#ifndef PBZIP_NO_LOADAVG
+	fprintf(stderr, " -l       : load average determines max number processors to use\n");
+#endif
 	fprintf(stderr, " -p#      : where # is the number of processors (default");
 #if defined(_SC_NPROCESSORS_ONLN) || defined(__APPLE__)
 	fprintf(stderr, ": autodetect)\n");
@@ -1444,6 +1461,11 @@ int main(int argc, char* argv[])
 	struct timeval tvStopTime;
 	#ifndef WIN32
 	struct timezone tz;
+	double loadAverage = 0.0;
+	double loadAvgArray[3];
+	int useLoadAverage = 0;
+	int numCPUtotal = 0;
+	int numCPUidle = 0;
 	#else
 	SYSTEMTIME systemtime;
 	LARGE_INTEGER filetime;
@@ -1506,6 +1528,9 @@ int main(int argc, char* argv[])
 	#elif defined(_SC_NPROCESSORS_ONLN)
 		numCPU = sysconf(_SC_NPROCESSORS_ONLN);
 	#endif
+	#ifndef WIN32
+	numCPUtotal = numCPU;
+	#endif
 
 	// parse command line switches
 	for (i=1; i < argc; i++)
@@ -1575,6 +1600,9 @@ int main(int argc, char* argv[])
 				case 'c': OutputStdOut = 1; keep = 1; break;
 				case 'f': force = 1; break;
 				case 'k': keep = 1; break;
+				#ifndef PBZIP_NO_LOADAVG
+				case 'l': useLoadAverage = 1; break;
+				#endif
 				case 'r': readEntireFile = 1; break;
 				case 't': testFile = 1; break;
 				case 'v': QuietMode = 0; break;
@@ -1620,6 +1648,38 @@ int main(int argc, char* argv[])
 				fprintf(stderr, " You will only able to compress files smaller than 2GB in size.\n\n");
 		}
 	}
+	
+	// Calculate number of processors to use based on load average if requested
+	#ifndef PBZIP_NO_LOADAVG
+	if (useLoadAverage == 1)
+	{
+		// get current load average
+		ret = getloadavg(loadAvgArray, 3);
+		if (ret != 3)
+		{
+			loadAverage = 0.0;
+			useLoadAverage = 0;
+			if (QuietMode != 1)
+				fprintf(stderr, " *WARNING: Could not get load average!  Using requested processors...\n");
+		}
+		else
+		{
+			#ifdef PBZIP_DEBUG
+			fprintf(stderr, "Load Avg1: %f  Avg5: %f  Avg15: %f\n", loadAvgArray[0], loadAvgArray[1], loadAvgArray[2]);
+			#endif
+			// use 1 min load average to adjust number of processors used
+			loadAverage = loadAvgArray[0];	// use [1] for 5 min average and [2] for 15 min average
+			// total number processors minus load average rounded up
+			numCPUidle = numCPUtotal - (int)(loadAverage + 0.5);
+			// if user asked for a specific # processors and they are idle, use all requested
+			// otherwise give them whatever idle processors are available
+			if (numCPUidle < numCPU)
+				numCPU = numCPUidle;
+			if (numCPU < 1)
+				numCPU = 1;
+		}
+	}
+	#endif
 
 	// setup signal handling
 	sigInFilename = NULL;
@@ -1636,6 +1696,10 @@ int main(int argc, char* argv[])
 		if (testFile != 1)
 		{
 			fprintf(stderr, "\n         # CPUs: %d\n", numCPU);
+			#ifndef PBZIP_NO_LOADAVG
+			if (useLoadAverage == 1)
+				fprintf(stderr, "   Load Average: %'.2f\n", loadAverage);
+			#endif
 			if (decompress != 1)
 			{
 				fprintf(stderr, " BWT Block Size: %d00k\n", BWTblockSize);
